@@ -14,10 +14,14 @@ import inspect
 import bisect
 import difflib
 import itertools
+import concurrent.futures as cf
 
 
 class _Sane:
-    VERSION = '6.1'
+    # Simplified versioning, MAJOR.MINOR.
+    # Breaking changes change MAJOR version, backwards compatible changes
+    # change MINOR version.
+    VERSION = '7.1'
 
     ### State ###
 
@@ -29,6 +33,7 @@ class _Sane:
         self.hooks = []
         self.recipe_calls = []
         self.force = False
+        self.threads = 1
 
     #### Reporting ####
 
@@ -112,9 +117,13 @@ class _Sane:
         exit(1)
 
     def set_verbose(self, verbose_level):
+        self.log(f'Verbose level set to {verbose_level}',
+                _Sane.VerboseLevel.DEBUG)
         self.verbose = verbose_level
 
     def set_ansi(self, use_ansi):
+        self.log(f'Using ANSI colors: {use_ansi}',
+                _Sane.VerboseLevel.DEBUG)
         self.use_ansi = use_ansi
 
     ### Dependency Graph ###
@@ -148,6 +157,10 @@ class _Sane:
         else:
             raise ValueError(f'Unimplemented type \'{type_}\'')
 
+    @staticmethod
+    def unique_name_is_recipe(unique_name):
+        return unique_name.startswith('recipe-')
+
     @staticmethod
     def split_unique_name(unique_name):
         if unique_name.startswith('hook-'):
@@ -220,6 +233,7 @@ class _Sane:
 
     def register_recipe(self, fn, name, hooks, recipe_deps,
                         hook_deps, conditions, info):
+        unique_name = _Sane.get_unique_name(name, _Sane.Node.RECIPE)
         # If we do not register the hooks in hook_deps, it may happen that
         # no recipe is ever registered with those hooks, leading to an
         # "unknown dependency" error. This is counter intuitive (if we depend
@@ -229,25 +243,23 @@ class _Sane:
                 ((True, x) for x in hooks),
                 ((False, x) for x in hook_deps))
         for connect, hook in iterator:
-            hook_node = \
+            hook_node = (
                 self.graph.setdefault(
                     _Sane.get_unique_name(hook, _Sane.Node.HOOK),
-                    _Sane.Node(_Sane.Node.HOOK))
+                    _Sane.Node(_Sane.Node.HOOK, connections=[])))
             if connect:
-                hook_node.connections.append(
-                    _Sane.get_unique_name(name, _Sane.Node.RECIPE))
+                hook_node.connections.append(unique_name)
             i = bisect.bisect_left(self.hooks, hook)
             if not (i < len(self.hooks) and self.hooks[i] == hook):
                 self.hooks.insert(i, hook)
 
-        connections = \
+        connections = (
             [_Sane.get_unique_name(recipe, _Sane.Node.RECIPE)
-             for recipe in recipe_deps] + \
+             for recipe in recipe_deps] + 
             [_Sane.get_unique_name(hook, _Sane.Node.HOOK)
-             for hook in hook_deps]
+             for hook in hook_deps])
 
-        unique_name = _Sane.get_unique_name(name, _Sane.Node.RECIPE)
-        self.graph[unique_name] = \
+        self.graph[unique_name] = (
             _Sane.Node(
                 _Sane.Node.RECIPE,
                 connections,
@@ -255,7 +267,7 @@ class _Sane:
                     'fn': fn,
                     'conditions': conditions,  # unique names
                     'info': info  # string
-                })
+                }))
 
         # Recipes are inserted sorted by name, so as to be able to
         # perform fuzzing later.
@@ -274,16 +286,13 @@ class _Sane:
             args, kwargs, frame = self.recipe_calls.pop()
 
             # "Unwrap" expected kwargs
-            name = kwargs['name']
-            hooks = kwargs['hooks']
-            recipe_deps = kwargs['recipe_deps']
-            hook_deps = kwargs['hook_deps']
-            conditions = kwargs['conditions']
-            info = kwargs['info']
-            fn = kwargs['fn']
-            del kwargs['fn'], kwargs['name'], kwargs['hooks'], \
-                kwargs['recipe_deps'], kwargs['hook_deps'], \
-                kwargs['conditions'], kwargs['info']
+            name = kwargs.pop('name', None)
+            hooks = kwargs.pop('hooks', None)
+            recipe_deps = kwargs.pop('recipe_deps', None)
+            hook_deps = kwargs.pop('hook_deps', None)
+            conditions = kwargs.pop('conditions', None)
+            info = kwargs.pop('info', None)
+            fn = kwargs.pop('fn', None)
 
             # Often, the user will wrongly decorate the function with `@recipe`,
             # instead of `@recipe()`, if they intend all the arguments to take
@@ -321,12 +330,8 @@ class _Sane:
                           'but may be ignored or fail in the future.')
                 conditions.append(
                     _Help.file_condition(
-                        sources=kwargs.get('file_deps', []),
-                        targets=kwargs.get('target_files', [])))
-                if file_deps_present:
-                    del kwargs['file_deps']
-                if target_files_present:
-                    del kwargs['target_files']
+                        sources=kwargs.pop('file_deps', []),
+                        targets=kwargs.pop('target_files', [])))
 
             # - Unknown keyword arguments
             if len(kwargs) > 0:
@@ -411,7 +416,7 @@ class _Sane:
                                f'as a hook dependency of recipe \'{name}\'.\n'
                                f'At {_Sane.trace(frame)}')
             self.log(
-                f'Hooks of recipe \'{name}\':\n{chr(10).join(hook_deps)}',
+                f'Hook deps of recipe \'{name}\':\n{chr(10).join(hook_deps)}',
                 _Sane.VerboseLevel.DEBUG)
 
             # - Condition dependency
@@ -433,6 +438,18 @@ class _Sane:
                 f'Conditions of recipe \'{name}\':\n'
                 f'{chr(10).join(str(x) for x in conditions)}',
                 _Sane.VerboseLevel.DEBUG)
+            
+            # - Hooks
+            for hook in hooks:
+                if type(hook) is not str:
+                    self.error(
+                            f'Hooks must be strings, but got hook \'{hook}\' '
+                            'as a hook of recipe \'{name}\'.\n'
+                            f'At {_Sane.trace(frame)}')
+            self.log(
+                f'Hooks of recipe \'{name}\':\n'
+                f'{chr(10).join(str(x) for x in hooks)}',
+                _Sane.VerboseLevel.DEBUG)
 
             # Register recipe
             self.log(
@@ -440,14 +457,24 @@ class _Sane:
                 _Sane.VerboseLevel.DEBUG)
             self.register_recipe(fn, name, hooks, recipe_deps, hook_deps,
                                  conditions, info)
-
             del frame
 
     ### Execution ###
 
     def set_force(self, force):
+        self.log(f'Forcing runs: {force}',
+                _Sane.VerboseLevel.DEBUG)
         self.force = force
 
+    def set_threads(self, threads):
+        self.log(f'Using {threads} threads.',
+                _Sane.VerboseLevel.DEBUG)
+        self.threads = threads
+
+    def run_recipe(self, fn, name):
+        self.log(f'Running recipe \'{name}\'.', _Sane.VerboseLevel.VERBOSE)
+        fn()
+
     def run_recipe_graph(self, recipe_name):
         """
         This is perhaps the most "sensitive" part of `sane`,
@@ -501,8 +528,8 @@ class _Sane:
         cases.
         """
         root_name = recipe_name
-        root_unique_name = \
-            _Sane.get_unique_name(root_name, _Sane.Node.RECIPE)
+        root_unique_name = (
+            _Sane.get_unique_name(root_name, _Sane.Node.RECIPE))
         root_node = self.graph[root_unique_name]
 
         active = []
@@ -603,26 +630,42 @@ class _Sane:
                 child_active = (child_node.is_always_active() or self.force)
                 backtrack_active.extend((child_active, False))
 
-        # Sort the active recipes by depth, and then by order in which they were
-        # encountered.
-        sort_key = sorted(range(len(active_ordering)),
-                          key=active_ordering.__getitem__)
-        active_in_order = map(lambda i: active[i], reversed(sort_key))
+        if len(active_ordering) > 0:
+            # Sort the active recipes by depth, and then by order in which they
+            # were encountered.
+            sort_key = sorted(range(len(active_ordering)),
+                              key=active_ordering.__getitem__)
+            active_in_order = map(lambda i: active[i], reversed(sort_key))
+            active_in_order = filter(
+                    lambda x: _Sane.unique_name_is_recipe(x[1]),
+                    active_in_order)
 
-        self.log('Finished building and sorting dependency tree of '
-                 f'\'{root_name}\'.\n'
-                 'Launching recipes.',
-                 _Sane.VerboseLevel.DEBUG)
+            depth, recipes = zip(*active_in_order)
+            recipes = list(recipes)
 
-        # Finally, run the recipes in order
-        # Note that some of the elements in `active` will be hooks.
-        for _depth, unique_name in active_in_order:
-            type_, name = _Sane.split_unique_name(unique_name)
-            if type_ != _Sane.Node.RECIPE:
-                continue
-            self.log(f'Running recipe \'{name}\'.', _Sane.VerboseLevel.VERBOSE)
-            fn = self.graph[unique_name].meta['fn']
-            fn()
+            idx_groups = itertools.groupby(range(len(depth)),
+                                           key=depth.__getitem__)
+
+            self.log('Finished building and sorting dependency tree of '
+                     f'\'{root_name}\'.\n'
+                     'Launching recipes.',
+                     _Sane.VerboseLevel.DEBUG)
+            
+            with cf.ThreadPoolExecutor(max_workers=self.threads) as pool:
+                for _, idx_group in idx_groups:
+                    group_recipes = map(recipes.__getitem__, idx_group)
+                    group_recipes = map(lambda name: (
+                                           self.graph[name].meta['fn'],
+                                           _Sane.split_unique_name(name)[1]),
+                                        group_recipes)
+                    futures = [pool.submit(self.run_recipe, *args) 
+                                for args in group_recipes]
+                    for result in cf.as_completed(futures):
+                        result.result()     # Block on this depth
+        else:
+            self.log(
+                f'Nothing to do for \'{root_name}\'',
+                _Sane.VerboseLevel.DEBUG)
 
 
 _stateful = _Sane()
@@ -798,17 +841,21 @@ def sane_run(default=None, cli=True):
                             help='List the defined recipes.')
         parser.add_argument('--no-ansi', action='store_true', default=False,
                             help='Disable ANSI color characters in logging.')
+        parser.add_argument('--threads', metavar='threads', type=int, default=1,
+                            help='The number of threads to run the tasks with.')
 
         args = parser.parse_args()
 
         _stateful.set_verbose(args.verbose)
         _stateful.set_ansi(not args.no_ansi)
         _stateful.set_force(args.force)
+        _stateful.set_threads(args.threads)
 
-        _stateful.log('Parsing registered `@recipe` decorations.',
-                      _Sane.VerboseLevel.DEBUG)
-        _stateful.parse_decorator_calls()
+    _stateful.log('Parsing registered `@recipe` decorations.',
+                  _Sane.VerboseLevel.DEBUG)
+    _stateful.parse_decorator_calls()
 
+    if cli:
         if args.list:
             _stateful.list_recipes()
             exit(0)
@@ -874,6 +921,14 @@ def sane_run(default=None, cli=True):
 
             recipe = default
     else:
+        # Type check `recipe`, even though it should be type checked by
+        # ArgumentParser, as programmatic calls can bypass this
+        if type(recipe) is not str:
+            _stateful.error(
+                'Not calling default recipe, and given recipe is not a '
+                'string.\n'
+                f'(Got \'{recipe}\'.)\n'
+                'Are you forgetting `cli=False`?')
         if not _stateful.recipe_exists(recipe):
             _stateful.error(
                 f'Given recipe \'{recipe}\' is not defined as a recipe.')
